﻿module pes.btree;

// ごくふつうの２分木


private import pes.exception;
private import std.string;
private import std.stdio;

class BinaryTree(K, V)
{
protected:
	alias BinaryTreeNode!(K, V)	BTNode;
	alias BinaryTreeNodePos!(K, V) BTNodePos;
	BTNode 	m_root;
	uint	m_length;
	
public:
	this()
	{
		m_root = new BTNode;
	}
	
	uint length()
	{
		return m_length;
	}
	
	BTNode root()
	{
		return m_root;
	}
	
	void set(K k, V v)
	{
		BTNode 		new_node		= new BTNode(k, v);
		BTNodePos 	node_pos	 	= root.searchPosition(k);
		
		if (*node_pos.pos !is null) {
			// 既に存在する
			new_node.right 	= (*node_pos.pos).right;
			new_node.left	= (*node_pos.pos).left;
		} else {
			++m_length;
		}
		
		// セット
		*node_pos.pos = new_node;
		
		new_node = null;
		node_pos = null;
	}
/*
	V get(K k)
	{
		BTNodePos node_pos = root.searchPosition(k);
		
		if (*node_pos.pos is null) {
			throw new NotFound(std.string.format("BTNode: Key(%s)",  k));
		}
		
		return (*node_pos.pos).value;
	}
*/	
	V get(K k)
	{
		return root.get(k);
	}
	
	void remove(K k)
	{
		BTNodePos	node_pos	= root.searchPosition(k);
		BTNode		node		= *node_pos.pos;
		
		if (node is null) {
			return;
		}
		
		if (node.left is null && node.right is null) {
			*node_pos.parent = null;
		}
		else if (node.left is null && node.right !is null) {
			*node_pos.parent = node.right;
			node.right = null;
		}
		else if (node.left !is null && node.right is null) {
			*node_pos.parent = node.left;
			node.left = null;
		}
		else { // left !is null && right !is null
			*node_pos.parent	= node.left;
			node.left = null;
			do {
				node_pos.parent = &(*node_pos.parent).right;
			} while (*node_pos.parent !is null);
			
			*node_pos.parent = node.right;
			node.right = null;
		}
		
		node = null;
		node_pos.parent = null;
		*node_pos.pos = null;
		node_pos = null;
		
		--m_length;
	}
	
	
	K[] keys()
	{
		K[] keys;
		uint size = m_length, i;
		
//		keys = new K[size];
		
		foreach (BTNode e; apply) {
//			keys[i++] = e.key;
			keys ~= e.key;
//			if (i >= size) {
//				break;
//			}
		}
		
		return keys;
	}
	
	V[] values()
	{
		V[] values;
		uint size = m_length, i;
		
//		values = new V[size];
		
		foreach (BTNode e; apply) {
//			values[i++] = e.value;
			values ~= e.value;
//			if (i >= size) {
//				break;
//			}
		}
		
		return values;
	}
	
	BTNode[] apply()
	{
		int result = 0;
		BTNode[] array;
		
		void treewalker(BTNode node)
		{
			do {
				array ~= node;
//				writef("%s %s\n", node.key, node.value);
				if (node.right !is null) {
					treewalker(node.right);
				}
				node = node.left;
			} while (node !is null);
		}
		
		// rootの右足から
		if (m_root.right !is null) {
			treewalker(m_root.right);
		}
		
		return array;
	}
	
	bool exist(K k)
	{
		BTNodePos	node_pos	= root.searchPosition(k);
		
		return *node_pos.pos is null ? false: true;
	}
}

class BinaryTreeNodePos(K, V)
{
	BinaryTreeNode!(K, V) *pos;
	BinaryTreeNode!(K, V) *parent;
	
	this(BinaryTreeNode!(K, V) *pos, BinaryTreeNode!(K, V) *parent)
	{
		this.pos = pos;
		this.parent = parent;
	}
}

class BinaryTreeNode(K, V)
{
public:
	BinaryTreeNode	right;
	BinaryTreeNode	left;
	
protected:
	alias	BinaryTreeNodePos!(K, V)		BTNodePos;
	K				m_key;
	V				m_value;
//	uint			m_hash;
	
	BTNodePos searchPosition_x(BinaryTreeNode *parent, K k)
	{
		int diff;
		
		// 比較 データ比較
		diff = typeid(K).compare(cast(void *)&m_key, cast(void *)&k);
		
		if (diff == 0) {
			return new BTNodePos(&this, parent);	// 一致
		}
		else if (diff > 0) {
			if (left is null) {
				return new BTNodePos(&left, &this);
			}
			
			return left.searchPosition_x(&left, k);
		}
		
		if (right is null) {
			return new BTNodePos(&right, &this);
		}
		
		return right.searchPosition_x(&right, k);
	}
	/*
	int opApply_x(int delegate(inout BinaryTreeNode) dg) 
	{
		int result = 0;
		
		if (left !is null) {
			result = left.opApply_x(dg);
			if (result) {
				return result;
			}
		}
		
		result = dg(this);
		if (result) {
			return result;
		}
		
		if (right !is null) {
			result = right.opApply_x(dg);
			if (result) {
				return result;
			}
		}
		
		return result;
	}
	*/
public:
	this()
	{
		
	}
	
	this(K key, V value)
	{
		m_key	= key;
		m_value	= value;
//		m_hash	= typeid(K).getHash(&m_key);
	}
	
	K key()
	{
		return m_key;
	}
	
//	uint getHash()
//	in {
//		assert(m_hash == typeid(K).getHash(&m_key));
//	}
//	body {
//		return m_hash;
//	}
	
	V value()
	{
		return m_value;
	}
	
	V get(K k)
	{
		BinaryTreeNode node;
		int diff;
		
		node = this.right;
		
		while (node !is null) {
			// 比較 データ比較
			diff = typeid(K).compare(cast(void *)&node.m_key, cast(void *)&k);
			if (diff == 0) {
				break;
			}
			if (diff > 0) {
				node = node.left;
			} else {
				node = node.right;
			}
		}
		if (node is null) {
			throw new NotFound(std.string.format("BinaryTree: Key(%s)",  k));
		}
		
		return node.value;
	}
	
	BTNodePos searchPosition(K k)
	{
		// rootからの検索
		// rootは右の足から木を作る
		if (right is null) {
			return new BTNodePos(&right, &this);
		}
		
		return right.searchPosition_x(&right, k);
	}
}

debug(pes_btree)
{
	import std.stdio;
	import std.string;
	import std.random;
	
	import pes.timer;
	int main()
	{
		auto Timer timer = new Timer;
		BinaryTree!(uint, char[]) btree = new BinaryTree!(uint, char[]);
		BinaryTree!(uint, char[]) btree2 = new BinaryTree!(uint, char[]);
		
		for (uint i = 0; i < 1000; ++i) {
			uint 	num = rand();
			char[]	str = toString(num);
			
			btree.set(num, str);
		}
		
		foreach (uint e; btree.keys) {
			if (e % 2) {
				btree.remove(e);
			}
		}
		
		int x;
		foreach (uint e; btree.keys) {
			btree2.set(e, btree.get(e));
			//writef("%d: %s - %s\n", x++, e, btree.get(e));
		}
		
		return 0;
	}
}
