﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.data.stt;

// 構造化テキストファイル


// かなり前に書いたもので、
// 自分でもよくわからないことになっているので
// そのうち絶対に書き直す!
// 全体的に統一性がない･･。つくかけだった気もする･･。



private import std.stdio;
private import std.path;
private import std.file;
private import std.stream;
private import std.mmfile;
private import std.string;
private import std.format;
private import std.conv;
private import std.ctype;
private import std.c.stdlib;
private import pes.variant;
private import pes.exception;

// 方向図
/* -----------------------------------------------------------------------------
 +   - STTList
 + 
 +         parent
 +           |
 +  prev--[NODE]--next
 +           |
 +        child
 +           |  isNode
 +         [NODE]--[NODE]--[NODE]--[NODE]--null
 +    isLeaf |       |       |       |   
 +        [VALUE] [VALUE]   null   [NODE]--[NODE]--null
 +         value                     |       | 
 +                                 null    [VALUE]
 +
 + <top>---------------------------------------------------------<bottom>
 + 
 +-----------------------------------------------------------------------------*/
abstract class STTList: Variant
{
protected:
	STTList	m_parent;
	STTList	m_next;
	STTList	m_prev;
	
	uint	m_line;
	char[]	m_name;
	
	static void notImplementedException(char[] name, char[] function_name)
	{
		throw new NotImplemented(name, function_name);
	}
	
	void notImplementedException(char[] function_name)
	{
		notImplementedException(
			format("line(%s): %s", m_line, this.classinfo.name), function_name
		);
	}
	
public:
	this(char[] name, uint line = 0)
	{
		m_name	= name.dup;
		m_line	= line;
	}
	
	abstract bool isNode();
	bool isLeaf()
	{
		return !isNode;
	}
	
	Variant value()
	{
		notImplementedException("value");
		return null;
	}
	
	
	override NumericType numericType()
	{
		return value.numericType;
	}
	
	override long toLong()
	{
		return value.toLong;
	}
	
	override real toReal()
	{
		return value.toReal;
	}
	
	override bool toBool()
	{
		return value.toBool;
	}
	
	override char[] toString()
	{
		return value.toString;
	}
	
	STTList child()
	{
		notImplementedException("child");
		return null;
	}
	
	STTList setChild(STTList child_node)
	{
		notImplementedException("setChild");
		return null;
	}
	
	char[] quoteString()
	{
		notImplementedException("quoteString");
		return null;
	}
	
	STTList setNext(STTList next_node)
	{
		m_next 				= next_node;
		next_node.m_prev 	= this;
		
		return next_node;
	}
	
	STTList insertNext(STTList next_node)
	{
		next_node.bottom.setNext(m_next);
		m_next 				= next_node;
		next_node.m_prev 	= this;
		
		return next_node;
	}
	
	STTList parent()
	{
		return m_parent;
	}
	
	STTList next()
	{
		return m_next;
	}
	
	STTList prev()
	{
		return m_prev;
	}
	
	STTList top()
	{
		STTList node = this;
		
		while (node.prev !is null) {
			node = node.prev;
		}
		
		return node;
	}
	
	STTList bottom()
	{
		STTList node = this;
		
		while (node.next !is null) {
			node = node.next;
		}
		
		return node;
	}
	
	STTList root()
	{
		STTList node = top;
		
		while (node.parent !is null) {
			node = node.parent.top;
		}
		
		return node;
	}
	
	char[] path()
	{
		char[] 	fullpath	= name;
		
		for (STTList node = top.parent; node !is null; node = node.top.parent) {
			fullpath = node.name ~ std.path.sep ~ fullpath;
		}
		
		return fullpath;
	}
	
	void sourcePosition(out char[] filename, out uint line)
	{
		filename = root.name.dup;
		line	 = m_line;
	}
	
	char[] name()
	{
		return m_name;
	}
	
	uint line()
	{
		return m_line;
	}
	
	int opApply(int delegate(inout STTList) dg)
	{
		int result = 0;
		
		for (STTList node = child; node !is null; node = node.next) {
			result = dg(node);
			if (result) {
				break;
			}
		}
		
		return result;
	}
	
	Variant opIndex(Variant name)
	{
		Variant [] array;
		char[] string_name	= name.toString;
		STTList node		= this;
		
		while (node.next !is null) {
			if (std.string.icmp(node.name, string_name) == 0) {
				array ~= node;
			}
			node = node.next;
		}
		
		return new VArray(array);
	}
	
	STTList opIndex(char[] name)
	{
		return findFirstNode(name);
	}
	
	STTList opIndex(int n)
	{
		STTList node = child;
		
		for (int i = 0; i < n; ++i) {
			if (node.next !is null) {
				node = node.next;
			} else {
				throw new NotFound("index " ~ std.string.toString(i));
			}
		}
		
		return node;
	}
	
	STTList[] findNodes(char[] name)
	{
		STTList[] results 	= null;
		STTList node 		= this;
		
		if (node is null) {
			throw new NotFound(name);
//			return results;
		}
		
		void find(STTList node) {
			foreach (STTList e; node) {
				if (name == e.name) {
					results ~= e;
				}
				if (e.isNode) {
					find(e);
				}
			}
		} find(node);
		
		return results;
	}
	
	STTList findFirstNode(char[] name)
	{
		STTList node 	= this;
		STTList result;
		
		if (node is null) {
			throw new NotFound(name);
///			return result;
		}
		
		STTList find(STTList node) {
			STTList hit;
			foreach (STTList e; node) {
				if (name == e.name) {
					return e;
				}
				if (e.isNode) {
					hit = find(e);
					if (hit !is null) {
						return hit;
					}
				}
			}
			
			return null;
		} 
		result = find(node);
		if (result is null) {
			throw new NotFound(name);
		}
		
		return result;
	}
}

/**
 *
 * STTNode: STTList
 *
 *
 */
class STTNode: STTList
{
protected:
	STTList	m_child;

public:
	this(char[] name, uint line = 0)
	{
		super(name, line);
	}
	
	TypeInfo type()
	{
		return typeid(STTNode);
	}
	
	bool isNode()
	{
		return true;
	}
	
	STTList child()
	{
		return m_child;
	}
	
	STTList setChild(STTList child_node)
	{
		m_child 			= child_node;
		child_node.m_parent = this;
		
		return child_node;
	}
}


/**
 *
 * STTLeaf: STTList
 *
 *
 */
class STTLeaf: STTList
{
protected:
	Variant	m_value;
	char[]	m_original;
	bool	m_has_change;
	
public:
	this(char[] name, char[] value, uint line = 0)
	{
		super(name, line);
		m_value 		= make(value);
		m_original		= value.dup;
		m_has_change	= false;
	}
	
	this(char[] name, Variant value, uint line = 0)
	{
		super(name, line);
		
		m_value 		= value;
		m_original		= value.toString.dup;
		m_has_change	= false;
	}
	
	TypeInfo type()
	{
		return typeid(STTLeaf);
	}
	
	bool isNode()
	{
		return false;
	}
	
	Variant value()
	{
		return m_value;
	}
	
	char[] quoteString()
	{
		char[] raw_value;
		char[] strip_value;
		
		
		if (m_has_change) {
			if (m_value.type is typeid(VString)) {
				raw_value	= m_value.toString;
				strip_value	= std.string.strip(raw_value);
				
				if (std.string.cmp(raw_value, strip_value) == 0) {
					return raw_value;
				} else {
					return "'"  ~ raw_value ~ "'";
				}
			}
			
			return m_value.toString;
		}
		raw_value 	= m_original;
		if (find(raw_value, '#') != -1) {
			return "'" ~ raw_value ~ "'";
		}
		strip_value	= std.string.strip(m_original);
		if (std.string.cmp(raw_value, strip_value) == 0) {
			return m_original;
		}
		
		return "'" ~ m_original ~ "'";
	}
	
	override Variant assign(Variant value)
	{
		m_value 		= value;
		m_has_change	= true;
		return this;
	}
}

/+-------------------------------------------------------------------+/
class STTSyntaxError: Exception
{
public:
	uint line;
	char[] name;
	
	this(char[] message, char[] ename, uint eline)
	{
		name = ename.dup;
		line = eline;
		super(name ~ "(" ~ std.conv.toString(line) ~ "):" ~ message);
	}
}

class STTConvError: Exception
{
public:
	uint line;
	char[] name;
	
	this (char[] message, char[] ename, uint eline)
	{
		line = eline;
		name = ename.dup;
		
		super(name ~ "(" ~ std.conv.toString(line) ~ "):" ~ message);
	}
}


class STTReader
{
protected:
	char[] 	m_stream_name;
	char[] 	m_indent_pattern;
	ulong	m_line;
	
	alias STTReader		ThisInstance;
	
public:
	static const int IGNORE_LINE = -1;
	
	this(char[] indent_pattern, char[] stream_name)
	{
		m_indent_pattern 	= indent_pattern.dup;
		m_stream_name		= stream_name.dup;
	}
	
	this()
	{}
	
	ThisInstance setIndentPattern(char[] indent_pattern)
	{
		m_indent_pattern = indent_pattern.dup;
		return this;
	}
	
	ThisInstance setStreamName(char[] stream_name)
	{
		m_stream_name = stream_name;
		return this;
	}
	
	STTList read(InputStream stream)
	{
		uint	i;
		int		level, current_level;
		STTList root, current_node, new_node;
		
		root			= new STTNode(m_stream_name); // root node
		current_level 	= 0;
		current_node	= root;
		
		foreach (char[] line; stream) {
			m_line++;
			
			level = parseLine(/* out */ new_node, line);
			if (level == IGNORE_LINE) {
				continue;
			}
			
			current_node = current_node.bottom;
			
			if (level == current_level) {
				current_node.setNext(new_node);
			}
			else if (level < current_level) {
				uint downhill_level = current_level - level;
				
				for (i = 0; i < downhill_level; ++i) {
					current_node = current_node.top.parent;
					if (current_node is null) {
						throw new STTSyntaxError("indent", m_stream_name, m_line);
					}
				}
				
				current_node.setNext(new_node);
			}
			else { // if (level > current_level) 
				if (current_node.isLeaf
					|| level - current_level != 1) // 
				{
					throw new STTSyntaxError("indent", m_stream_name, m_line);
				}
				
				current_node.setChild(new_node);
			}
			
			current_node	= new_node;
			current_level 	= level;
		}
		
		return root;
	}
	
	protected int parseLine(out STTList new_node, char[] line)
	{
		enum STTTokenType {
			NULL,
			COMMENT,
			NODE,
			LEAF,
		}
		STTTokenType	token;
		ulong			i, j;
		int				level;
		char[]			name, value;
		
		if (line.length == 0) {
			return IGNORE_LINE;
		}
		
		for (i = 0; i < line.length; ++i) {
			if (line[i] != '\t' && line[i] != ' ') {
				break;
			}
		}
		if (i == 0) {
			level = 1;
		} else {
			level = parseIndent(line[0 .. i], m_indent_pattern); // exception
		}
		
	__SKIP_NAME:
		for (; i < line.length; ++i) {
			switch (line[i]) {
			case ':':
				token = STTTokenType.NODE;
				break __SKIP_NAME;
			case '=':
				token = STTTokenType.LEAF;
				break __SKIP_NAME;
			case '#':
				token = STTTokenType.COMMENT;
				break __SKIP_NAME;
			default:
				break;
			}
		}
		if (i == line.length || token == STTTokenType.COMMENT) {
			return IGNORE_LINE;
		}
		name = std.string.strip(line[0 .. i]);
		if (name.length == 0) {
			name = "_";
		}
		++i;
		
		switch (token) {
			
		case STTTokenType.NODE:
			new_node = new STTNode(name, m_line);
			break;
			
		case STTTokenType.LEAF:
			value = std.string.strip(line[i .. length]);
			if (value.length == 0) {
				value = null;
			} else if (value[0] == '\'') {
				char[] 	esc_value;
				
				for (j = 1; j < value.length; ++j) {
					if (value[j] == '\'') {
						if (j + 1 < value.length && value[j + 1] == '\'') {
							esc_value ~= '\'';
							++j;
						} else {
							break;
						}
					} else {
						esc_value ~= value[j];
					}
				}
				if (esc_value.length == 0) {
					value = null;
				} else {
					value = esc_value;
				}
			} else {
				int comment_pos = std.string.find(value, '#');
				
				if (comment_pos == 0) {
					value = null;
				} else if (comment_pos > 0) {
					value = value[0 .. comment_pos];
				}
			}
			
			new_node = new STTLeaf(name, value, m_line);
			break;
		}
		
		return level;
	}
	
	protected int parseIndent(char[] indent, char[] indent_pattern)
	in {
		foreach (char c; indent_pattern) {
			if (c != 't' && c != 'T' && c != 's' && c != 'S') {
				assert(0);
			}
		}
	}
	out (result) {
		assert(result > 0);
	}
	body
	{
		uint i, j, level = 1;
		
		if (indent.length == 0) {
			return level;
		}
		
		for (j = 0, i = 0; j < indent.length; ++j, ++i) {
			if (i == indent_pattern.length) {
				i = 0;
				if (++level < 0) {
					// overflow
				}
			}
			if (indent_pattern[i] == 's' || indent_pattern[i] == 'S') {
				if (indent[j] != ' ') {
					throw new STTSyntaxError("indent", m_stream_name, m_line);
				}
			} else if (indent_pattern[i] == 't' || indent_pattern[i] == 'T') {
				if (indent[j] != '\t') {
					throw new STTSyntaxError("indent", m_stream_name, m_line);
				}
			}
		}
		
		if (i == indent_pattern.length) {
			if (++level < 0) {
				// overflow
			}
		} else {
			throw new STTSyntaxError("indent", m_stream_name, m_line);
		}
		
		return level;
	}
}

void outputSTT(STTList root, char[] ipat, OutputStream stream)
{
	char[] indent_string = indentString(ipat);
	
	void recursionOut(STTList root, int indent) {
		foreach (STTList node; root) {
			for (int i = 0; i < indent; ++i) {
				stream.writeString(indent_string);
			}
			if (node.isNode) {
				stream.writeLine(node.name ~ ":");
				recursionOut(node, indent + 1);
			} else {
				if (node.name.length == 0) {
					stream.writeLine("= " ~ (cast(STTLeaf)node).quoteString);
				} else {
					stream.writeLine(node.name ~ " = " ~ (cast(STTLeaf)node).quoteString);
				}
			}
		}
	} recursionOut(root, 0);
}

char[] indentString(char[] ipat)
in {
	assert(ipat.length != 0);
	foreach (char c; ipat) {
		if (c != 't' && c != 'T' && c != 's' && c != 'S') {
			assert(0);
		}
	}
}
body {
	char[] indent;
	
	foreach (char c; ipat) {
		if (c == 's' || c == 'S') {
			indent ~= " ";
		} else {
			indent ~= "\t";
		}
	}
	
	return indent;
}

char[] textSTT(STTList root, char[] indent_pattern)
{
	char[] text;
	char[] indent_string = indentString(indent_pattern);
	
	void recurrentJoin(STTList root, int indent)
	{
		foreach (STTList node; root) {
			for (int i = 0; i < indent; ++i) {
				text ~= indent_string;
			}
			if (node.isNode) {
				text ~= node.name ~ ":\n";
				recurrentJoin(node, indent + 1);
			} else {
				if (node.name.length == 0) {
					text ~= "= " ~ node.quoteString ~ "\n";
				} else {
					text ~= node.name ~ " = " ~ node.quoteString ~ "\n";
				}
			}
		}
	} recurrentJoin(root, 0);
	
	return text;
}

void saveSTT(STTList root, char[] ipat, char filepath[], FileMode mode = FileMode.OutNew)
{
	auto File	stream	= new File(filepath, mode);
	
	outputSTT(root, ipat, stream);
}

static STTList loadSTT(char[] filepath, char[] indent_pattern)
{
//	auto MmFile 		mmfile 	= new MmFile(filepath, MmFile.Mode.Read, 0, null);
//	auto MmFileStream	stream 	= new MmFileStream(mmfile);
	File				stream  = new File(filepath, FileMode.In);
	STTReader 			reader	= new STTReader;
	STTList				root;
	
	{
		scope(exit) stream.close();
		
		root = 			reader
					.setIndentPattern(indent_pattern)
					.setStreamName(filepath)
				.read(stream); //;-)
	}
	return root;
}

void printSTT(STTList root, char[] ipat)
{
	char[] indent_string = indentString(ipat);
	
	void recursionJoin(STTList root, int indent)
	{
		foreach (STTList node; root) {
			char[] filename;
			uint line;
			node.sourcePosition(filename, line);
			writef("%s(%06s)| ", filename, line);
			for (int i = 0; i < indent; ++i) {
				writef("%s", indent_string);
			}
			if (node.isNode) {
				writef("%s:\n", node.name);
				recursionJoin(node, indent + 1);
			} else {
				writef("%s = %s\n", node.name, node.quoteString);
			}
		}
	} recursionJoin(root, 0);
}

debug(pes_stt)
{
	import std.stdio;
	import std.path;
	import std.date;
	
	class Timer
	{
		static long getCount()
		{
/*			asm {
				naked	;
				rdtsc	;
				ret		;
			}*/
			return getUTCtime();
		}
		
		ulong starttime;
		
		void start()
		{
			starttime = getCount();
		}
		ulong stop()
		{
			return getCount() - starttime;
		}
	}
	
	int main(char[][] argv)
	{
		Timer	timer = new Timer;
		STTList root;
		STTReader reader;
		
		if (argv.length != 2) {
			writef("usage: %s sttfile\n", getBaseName(argv[0]));
			return -1;
		}
		timer.start();
		try {
			root = loadSTT(argv[1], "t");
			printSTT(root, "ssss");
		} catch (Exception e) {
			writef("Exception - %s\n", e.toString);
		}
		writef("time: %s\n", timer.stop);
		return 0;
	}
}

