/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.net.dns.name;

private import std.string;
private import pes.net.exception;

class DomainName
{
protected:
	char [][] 	name;
	size_t		original_length;
	
public:
	this(char [] domain)
	{
		name = split(domain, ".");
	}
	
	this(void [] header_raw, void [] data_raw)
	{
		name 			= unpack(header_raw, data_raw);
		original_length	= length(cast(ubyte [])data_raw);
	}
	
	size_t originalSize()
	in {
		assert(original_length != 0);
	}
	body {
		return original_length;
	}
	
	char [][] getName()
	{
		char [][] dupname;
		
		foreach (char [] elem; name) {
			dupname ~= elem.dup;
		}
		
		return dupname;
	}
	
	void [] pack()
	{
		ubyte [] packet;
		
		foreach (char [] elem; name) {
			packet ~= cast(ubyte)(elem.length);
			packet ~= cast(ubyte [])elem;
		}
		packet ~= cast(ubyte)0;
		
		return packet;
	}
	
	override char [] toString()
	{
		return join(name, ".");
	}
	
	static char [][] unpack(void [] header, void [] data)
	{
		char  [][] name;
		ubyte [] datap = cast(ubyte [])data;
		ubyte [] hdrsp = cast(ubyte [])header;
		
		size_t offset = 0, point;
		
		do {
			if (isPointer(datap[offset])) {
				point = pointer(datap[offset .. offset + 2]);
				name ~= unpack(hdrsp[point .. hdrsp.length], hdrsp);
				offset += ushort.sizeof;
			} else {
				name ~= label(datap, offset);
				offset += labelLength(datap, offset) + 1;
			}
		} while (datap[offset] != 0);
		
		return name;
	}
	
	static size_t length(void [] data_raw)
	{
		ubyte [] data = cast(ubyte [])data_raw;
		size_t offset = 0;
		
		do {
			if (isPointer(data[offset])) {
				offset += ushort.sizeof;
			} else {
				offset += data[offset] + 1;
			}
		} while (data[offset] != 0);
		
		return offset;
	}
	
	static bool isPointer(ubyte point)
	{
		if ((point & 0xC0) && ((point & 0xC0) != 0xC0)) {
			throw new ProtocolException("DNS", "Illegal label offset");
		}
		
		return (point & 0xC0) == 0xC0;
	}
	
	static bool isLabel(ubyte point) 
	{
		return !isPointer(point);
	}
	
	static size_t pointer(ubyte [] data, size_t offset = 0)
	in {
		assert(isPointer(data[offset]) == true);
	}
	body {
		return cast(size_t)((data[offset] & ~0xC0) << 8 | data[offset + 1]);
	}
	
	static ubyte labelLength(ubyte [] data, size_t offset = 0)
	in {
		assert(isLabel(data[offset]) == true);
	}
	body {
		return (data[offset] & ~0xC0);
	}
	
	static char [] label(ubyte [] data, size_t offset = 0)
	in {
		assert(isLabel(data[offset]) == true);
	}
	body {
		size_t bpos	= offset + 1;
		return cast(char [])data[bpos .. bpos + labelLength(data, offset)].dup;
	}
}

debug(pes_net_dns_name)
{
	import std.stdio;
	import std.string;
	
	int main()
	{
		static ubyte [] head = 
		[
			0x03, '1', '2', '3', 0x02, 'a', 'd', 0x02, 'j', 'p', 0x00,
			0x04, 'h', 'o', 'g', 'e', 0xC0, 0x04, 0x3, 'c', 'o', 'm', 0x00
		];
		ubyte [] data = head[11 .. head.length];
		
		DomainName domain = new DomainName(head, data);
		writef("%d %s\n", domain.originalSize, domain.toString);
		
		DomainName domain2 = new DomainName("www.yahoo.co.jp");
		writef("%s\n", cast(ubyte [])domain2.pack());
		return 0;
	}
}

