/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.queue;

// FIFO Queue


class Queue(T)
{
protected:
	alias	QueueNode!(T)	NODE_TYPE;
	
	uint 		m_count;
	NODE_TYPE	m_top;
	NODE_TYPE	m_bottom;
	
	class QueueBoundsError :Error
	{
		this()
		{
			super("pes.queue: QueueBoundsError!!");
		}
	}
	
public:
	this() { };
	
	bool empty()
	{
		return m_count == 0 ? true: false;
	}
	
	void push(T data)
	{
		if (m_top is null) {
			m_top 		= new NODE_TYPE(data);
			m_bottom	= m_top;
		} else {
			m_top = m_top.push(data);
		}
		++m_count;
	}
	
	T shift()
	{
		T data;
		
		if (m_count == 0) {
			throw new QueueBoundsError;
		}
		if (--m_count == 0) {
			data = m_top.data.typed_data;
			m_top = null;
		} else {
			data = m_top.shift;
		}
		
		return data;
	}
	
	uint length()
	{
		return m_count;
	}
}

class QueueNode(T)
{
protected:
	struct TStruct
	{
		T typed_data;
	}
	
public:
	TStruct 			data;
	QueueNode			next;
	QueueNode			prev;
	
	this(T data)
	{
		this.data.typed_data = data;
	}
	
	QueueNode push(T data)
	{
		QueueNode new_node = new QueueNode(data);
		
		new_node.prev = this;
		next = new_node;
		
		return new_node;
	}
	
	T shift()
	{
		QueueNode	 	node = this;
		T				data;
		
		while (node.prev !is null) {
			node = node.prev;
		}
		if (node.next !is null) {
			node.next.prev = null;
		}
		node.next = null;
		node.prev = null;
		
		data = node.data.typed_data;
		
		node.data = node.data.init;
		
		return data;
	}
}

debug(pes_queue)
{
	import std.stdio;
	
	class hoge {
		int data;
	}
	
	int main()
	{
		Queue!(hoge) queue = new Queue!(hoge);
		int i;
		
		for (i = 0; i < 10; ++i) {
			hoge d = new hoge;
			d.data = i;
			queue.push(d);
		}
		
		for (i = 0; i < 10; ++i) {
			writef("%d\n", queue.shift.data);
		}
		
		return 0;
	}
}
