﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module task.connectiontask;

// コネクション管理タスク



private import	std.thread,
				std.socket,
				std.stream,
				std.random;

private import	pes.system.all,
				pes.net.winny.all;

private import common.environment;
private import common.task;
private import common.packet;

debug(ui_task)
{
	import std.stdio;
	import std.cstream;
	import pes.prompt;
	import pes.dump;
}

debug(query_task)
{
	import pes.prompt;
}

class ConnectionTask: TaskBase
{
protected:
	ConnectionTable		m_upstream_connection;	// 上流検索リンク
	ConnectionTable		m_downstream_connection;// 下流検索リンク
	ConnectionTable		m_ul_data_connection;	// ULデータ転送リンク
	
	override TaskId taskId()
	{
		return TaskId.CONNECTION;
	}
	
	override void initialize()
	{
		// 優先順位をあげる
//		setPriority(PRIORITY.INCREASE);
		gcLimit = gcLimit * 2;
	}
	
	override void terminate()
	{
		// コネクションを全て切断後クリア
		m_upstream_connection.clear();
		m_downstream_connection.clear();
		m_ul_data_connection.clear();
	}
	
	override ReturnCode control()
	{
		ReturnCode result;
		
		while (true) {
			result = control_x();
			if (result != ReturnCode.SUCCESS) {
				break;
			}
			delayTask(10);
		}
		
		return result;
	}
	
	ReturnCode control_x()
	{
		Mail mail;
		debug(connection_task_trace) writef("%s\n", "control_x");
		
		/* コネクションタスクメインループ */
		if (!pollRequest(10)) {
			try {
				/* 接続済みソケットをポーリング */
				pollConnections();
			} catch(SocketException e) {
				debug(connection_task) mbswritef(derr, "Connection: Exception %s\n", e.toString);
			}
		} else {
			mail = fetchRequest();
			
			switch (mail.messageId) {
			case MessageId.TERMINATE:
				return ReturnCode.TERMINATE;	/* 終了	*/
			case MessageId.REBOOT:
				return ReturnCode.CONTINUE;
				
			case MessageId.SEND_COMMAND:
				sendCommand(mail.data);
				fullCollect(); // super.fullCollect GC
				break;
			default:
				break;
			}
		}
		
		return ReturnCode.SUCCESS;
	}
	
	// 他タスクからのコマンドを転送
	void sendCommand(Object opt)
	{
		debug(connection_task_trace) writef("%s\n", "sendCommand");
		
		m_upstream_connection.signal();
		m_downstream_connection.signal();
		m_ul_data_connection.signal();
		try {
			sendCommand_x(opt);
		} finally {
			m_upstream_connection.release();
			m_downstream_connection.release();
			m_ul_data_connection.release();
		}
	}
	
	void sendCommand_x(Object opt)
	{
		debug(connection_task_trace) writef("%s\n", "sendCommand_x");
		CommandSendOption 	option;
		NyCommand			command;
		
		try {
			unboxCommand(opt, option, command);
		} catch (Exception e) {
			debug(nodetask) mbswritef(derr, "Connection: exception %s\n", e.toString);
			return;
		}
//		debug(query_task) mbswritef(derr,
//			"Connection: コマンド転送[type:%d, to:%s, command: %s] \n",
//			cast(int)option.target,
//			option.sendTo is null ? "none":option.sendTo.toString,
//			command.typeInfo.toString
//		);
		
	_SWITCH:
		switch (option.target) {
		case CommandSendOption.Target.DownstreamConnection:
			if (m_downstream_connection.length == 0) {
				break;
			}
			if (option.sendTo is null) {
				// 指定なし
				Connection conn = m_downstream_connection.values
									[rand() % m_downstream_connection.length];
				try {
					conn.connection.timeout(config.timeout, 0);
					conn.connection.sendCommand(command);
				} catch (SocketException se) {
					m_downstream_connection.remove(conn.aaKey);
					break;
				}
				
			} else {
				// 指定あり
				
				foreach (Connection conn; m_downstream_connection.values) {
					if (compareNode(conn.nodeDetails.inetAddress, option.sendTo) 
						|| compareNode(conn.nodeDetails.reportAddress, option.sendTo)) 
					{
						try {
							conn.connection.timeout(config.timeout, 0);
							conn.connection.sendCommand(command);
						} catch (SocketException se) {
							m_downstream_connection.remove(conn.aaKey);
							break;
						}
						break;
					}
				}
			}
			break;
		case CommandSendOption.Target.DownstreamConnectionAll:
			
			foreach (Connection conn; m_downstream_connection.values) {
				try {
					conn.connection.timeout(config.timeout, 0);
					conn.connection.sendCommand(command);
				} catch (SocketException se) {
					m_downstream_connection.remove(conn.aaKey);
					break;
				}
			}
			
			break;
		case CommandSendOption.Target.UpstreamConnection:
			if (m_upstream_connection.length == 0) {
				break;
			}
			if (option.sendTo is null) {
				// 指定なし
				Connection conn = m_upstream_connection
									.values[rand() % m_upstream_connection.length];
				
				try {
					conn.connection.timeout(config.timeout, 0);
					conn.connection.sendCommand(command);
				} catch (SocketException se) {
					m_upstream_connection.remove(conn.aaKey);
					break;
				}
			} else {
				
				foreach (Connection conn; m_upstream_connection.values) {
					if (compareNode(conn.nodeDetails.inetAddress, option.sendTo) 
						|| compareNode(conn.nodeDetails.reportAddress, option.sendTo)) 
					{
						try {
							debug(query_task_d) mbswritef(derr, "Connection: コマンド転送成功[%s]\n",
																		command.classinfo.name);
							conn.connection.timeout(config.timeout, 0);
							conn.connection.sendCommand(command);
						} catch (SocketException se) {
							m_upstream_connection.remove(conn.aaKey);
							break;
						}
						break;
					}
				}
			}
			break;
		case CommandSendOption.Target.UpstreamConnectionAll:
			foreach (Connection conn; m_upstream_connection.values) {
				try {
					conn.connection.timeout(config.timeout, 0);
					conn.connection.sendCommand(command);
				} catch (SocketException se) {
					m_upstream_connection.remove(conn.aaKey);
					break;
				}
			}
			
			break;
		case CommandSendOption.Target.AllConnection:
			if (option.sendTo is null) {
				// 指定なし
				int which;
				Connection conn;
				
				if (m_upstream_connection.length == 0
					&& m_downstream_connection.length == 0)
				{
					break;
				} else if (m_upstream_connection.length != 0
							&& m_downstream_connection.length == 0)
				{
					which = 1;
				} else if (m_upstream_connection.length == 0
							&& m_downstream_connection.length != 0)
				{
					which = 0;
				} else {
					which = rand() % 2;
				}
				
				if (which) {
					conn = m_upstream_connection
								.values[rand() % m_upstream_connection.length];
				} else {
					conn = m_downstream_connection
							.values[rand() % m_downstream_connection.length];
				}
				
				try {
					conn.connection.timeout(config.timeout, 0);
					conn.connection.sendCommand(command);
				} catch (SocketException se) {
					if (which) {
						m_upstream_connection.remove(conn.aaKey);
					} else {
						m_downstream_connection.remove(conn.aaKey);
					}
					break;
				}
				
			} else {
				foreach (Connection conn; m_upstream_connection.values) {
					if (compareNode(conn.nodeDetails.inetAddress, option.sendTo) 
						|| compareNode(conn.nodeDetails.reportAddress, option.sendTo)) 
					{
						try {
							conn.connection.timeout(config.timeout, 0);
							conn.connection.sendCommand(command);
							debug(query_task_d) mbswritef(derr, "Connection: コマンド転送成功[%s]\n",
																		command.classinfo.name);
						} catch (SocketException se) {
							m_upstream_connection.remove(conn.aaKey);
							break;
						}
						break _SWITCH;
					}
				}
				foreach (Connection conn; m_downstream_connection.values) {
					if (compareNode(conn.nodeDetails.inetAddress, option.sendTo) 
						|| compareNode(conn.nodeDetails.reportAddress, option.sendTo)) 
					{
						try {
							conn.connection.timeout(config.timeout, 0);
							conn.connection.sendCommand(command);
							debug(query_task_d) mbswritef(derr, "Connection: コマンド転送成功[%s]\n",
																		command.classinfo.name);
						} catch (SocketException ) {
							m_downstream_connection.remove(conn.aaKey);
							break;
						}
						break _SWITCH;
					}
				}
			}
			break;
		case CommandSendOption.Target.AllConnectionAll:
			foreach (Connection conn; m_downstream_connection.values) {
				try {
					conn.connection.timeout(config.timeout, 0);
					conn.connection.sendCommand(command);
				} catch (SocketException ) {
					m_downstream_connection.remove(conn.aaKey);
				}
			}
			foreach (Connection conn; m_upstream_connection.values) {
				try {
					conn.connection.timeout(config.timeout, 0);
					conn.connection.sendCommand(command);
				} catch (SocketException ) {
					m_upstream_connection.remove(conn.aaKey);
				}
			}
			break;
		case CommandSendOption.Target.NewLink:
		default:
			break;
		}
		
		
		command.clearBuffer();
		command.destroy;
		delete command;
		command = null;
	}
	
	void pollConnections()
	{
		debug(connection_task_trace) writef("%s %d\n", "pollConnection", 0);
		NyRawCommand command;
		
		// こいつは一度にselectするべきなのでライブラリ側をなおさなあかん
		
		m_upstream_connection.signal();
		try {
			foreach (Connection conn; m_upstream_connection.values) {
				// チェックのみが目的のためすごく気を短くしとく
				conn.connection.timeout(0, 1000);
				debug(connection_task_trace) writef("%s %d\n", "pollConnection", 1);
				
				try {
					conn.connection.poll(NyConnection.Poll.READABLE);
				} catch (ConnectionTimeOut) {
					continue;
				} catch (SocketException e) {
					debug(connection_task) mbswritef(derr, "Connection: 検索リンク(上流)切断(%s:%s)\n",
															conn.connection.address, e.toString);
					m_upstream_connection.remove(conn.aaKey);
					continue;
				}
				debug(connection_task_trace) writef("%s %d\n", "pollConnection", 2);
				
				// インターネット側だとデータが落ちることがあるのでブロック覚悟で…
//				conn.connection.timeout(config.timeout, 0);
				
				try {
					command = conn.connection.receiveCommand();
					if (routingCommand(command, conn.nodeDetails, conn) == false) {
						debug(connection_task) 
						mbswritef(derr, "Connection: 検索リンク(上流)切断(%s)\n",
										conn.connection.address);
						m_upstream_connection.remove(conn.aaKey);
					}
					command = null;
				} catch (ConnectionTimeOut ) {
					// nothing
					continue;
				} catch (SocketException e) {
					debug(connection_task) mbswritef(derr, "Connection: 検索リンク(上流)切断(%s:%s)\n",
															conn.connection.address, e.toString);
					m_upstream_connection.remove(conn.aaKey);
					continue;
				}
			}
		} finally {
			m_upstream_connection.release();
		}
		
		debug(connection_task_trace) writef("%s %d\n", "pollConnection", 5);
		delayTask(5);
		
		m_downstream_connection.signal();
		try {
			foreach (Connection conn; m_downstream_connection.values) {
				// チェックのみが目的のためすごく気を短くしとく
				conn.connection.timeout(0, 1000);
				debug(connection_task_trace) writef("%s %d\n", "pollConnection", 6);
				
				try {
					conn.connection.poll(NyConnection.Poll.READABLE);
				} catch (ConnectionTimeOut) {
					continue;
				} catch (SocketException e) {
//					debug(connection_task) mbswritef(derr, "Connection: 検索リンク(上流)切断(%s:%s)\n",
//															conn.connection.address, e.toString);
					m_downstream_connection.remove(conn.aaKey);
					continue;
				}
				debug(connection_task_trace) writef("%s %d\n", "pollConnection", 7);
				
//				conn.connection.timeout(config.timeout, 0);
				
				try {
					command = conn.connection.receiveCommand();
					if (routingCommand(command, conn.nodeDetails, conn) == false) {
//						debug(connection_task) 
//							mbswritef(derr, "Connection: 検索リンク(下流)切断(%s)\n",
//											conn.connection.address);
						m_downstream_connection.remove(conn.aaKey);
					}
					command = null;
				} catch (ConnectionTimeOut ) {
					continue;
				} catch (SocketException ) {
//					debug(connection_task) mbswritef(derr, "Connection: 検索リンク(下流)切断(%s)\n", conn.connection.address);
					m_downstream_connection.remove(conn.aaKey);
					continue;
				}
			}
		} finally {
			m_downstream_connection.release();
		}
		
		debug(connection_task_trace) writef("%s %d\n", "pollConnection", 10);
		delayTask(5);
		
		m_ul_data_connection.signal();
		try {
			foreach (Connection conn; m_ul_data_connection.values) {
				// チェックのみが目的のためすごく気を短くしとく
				if (conn.isLock) {
					continue; // lock中
				}
				conn.connection.timeout(0, 1000);
				debug(connection_task_trace) writef("%s %d\n", "pollConnection", 11);
					
				try {
					conn.connection.poll(NyConnection.Poll.READABLE);
				} catch (ConnectionTimeOut) {
					continue;
				} catch (SocketException e) {
//					debug(connection_task) mbswritef(derr, "Connection: 転送リンク切断(%s:%s)\n",
//															conn.connection.address, e.toString);
					m_ul_data_connection.remove(conn.aaKey);
					continue;
				}
				debug(connection_task_trace) writef("%s %d\n", "pollConnection", 12);
				
//				conn.connection.timeout(config.timeout, 0);
				
				try {
					command = conn.connection.receiveCommand();
					if (routingCommand(command, conn.nodeDetails, conn) == false) {
//						debug(connection_task) 
//							mbswritef(derr, "Connection: 転送リンク切断(%s)\n",
//											conn.connection.address);
						m_ul_data_connection.remove(conn.aaKey);
					}
					command = null;
				} catch (ConnectionTimeOut ) {
					m_ul_data_connection.remove(conn.aaKey);	// ここは切断
					continue;
				} catch (SocketException ) {
//					debug(connection_task) mbswritef(derr,
//								"Connection: 転送リンク切断(%s)\n", conn.connection.address);
					m_ul_data_connection.remove(conn.aaKey);
					continue;
				}
			}
		} finally {
			m_ul_data_connection.release();
		}
		command = null;
	}
	
	/* 	自分で処理するものはする							*/
	/*	しないものはルーティングタスクにルーティングさせる 	*/
	bool routingCommand(NyRawCommand command, NodeDetails node, Connection conn)
	{
		debug(connection_task_trace) writef("%s\n", "routingCommand");
//		debug(connection_task) 
//			mbswritef(derr, "Connection: 受信 %s\n", command.toString);
		
		ReceivedCommand recv_command;
		bool keep_alive = false;
		
		switch (command.commandCode) {
		case NyCommandCode.CLOSE:
			break;
		case NyCommandCode.CONNECTED_LIMITATION:
			break;
		case NyCommandCode.SLOW_RATE:
			debug(connection_task) 
				mbswritef(derr, "Connection: 低速回線警告\n");
			log("SLOW_RATE command was received");
			break;
		case NyCommandCode.LOW_VERSION:
			debug(connection_task) 
				mbswritef(derr, "Connection: 低バージョン警告\n");
			log("LOW_VERSION command was received");
			break;
		case NyCommandCode.REJECT:
			debug(connection_task) 
				mbswritef(derr, "Connection: 無視ノード警告\n");
			log("REJECT command was received");
			break;
		case NyCommandCode.LIAR:
			debug(connection_task) 
				mbswritef(derr, "Connection: 捏造警告\n");
			log("LIAR command was received");
			break;
			
		case NyCommandCode.WRONG_LISTENING_PORT:
			/* 警告送信後 切断 */
			debug(connection_task) 
				mbswritef(derr, "Connection: ポート警告\n");
			log("WRONG_LISTENING_PORT command was received");
			break;
		case NyCommandCode.FILE_REQUEST:
			if (conn !is null) {
				conn.lock();
			} else {
				assert(0); // bug
			}
			debug(download) 
				mbswritef(derr, "Connection: ダウンロードリクエスト受信\n");
			
			// next
		default:
			// OK, ルーティングんに任せる
			recv_command			= new ReceivedCommand;
			recv_command.command 	= command;
			recv_command.node		= node;
			recv_command.connection	= conn;
			
			sendRequestTo(MailId.CONTROL_REQUEST, 
						Mail.make(MessageId.ROUTING, 
							MailId.NODE_REQUEST, 0, recv_command));
			keep_alive = true;
			break;
		}
		if (!keep_alive) {
			command.clearBuffer();
		}
		recv_command = null;
		
		fullCollect(); // super.fullCollect
		
		return keep_alive;
	}
	
public:
	this(Environment environ, TaskResource resource)
	{
		super(MailId.CONNECTION_REQUEST,
				MailId.CONNECTION_RESPONSE, environ, resource);
		
		// ローカルに取っとく
		m_upstream_connection	= environ.upstreamConnection;
		m_downstream_connection	= environ.downstreamConnection;
		m_ul_data_connection	= environ.ulDataConnection;
	}
}

