﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module task.controltask;

// ルーティング &
/* ワーカーコントロールタスク	*/

private import 	std.thread,
				std.c.time,
				std.socket,
				std.file;
private import 	pes.system.all,
				pes.net.winny.all,
				pes.conv,
				pes.arraybox,
				pes.exception,
				pes.queue,
				pes.file;

private import common.task;
private import common.task;
private import common.environment;
private import common.packet;
private import common.cacheresolve;
private import worker.uploader;
private import worker.cachedecoder;
private import worker.cacheencoder;
private import worker.cachebitmap;
private import task.timertask;

debug(control_task)
{
	import pes.prompt;
}

ulong toHostKey(uint address, uint port)
{
	ulong key = address;
	
	key <<= 32;
	
	return (key|port);
}

class ControlTask: TaskBase
{
protected:
	const uint WORKER_SCHD_TICK	= 0;
	Queue!(char[])			m_upload_files;
	Queue!(DownloadItem)	m_decode_items;
	
	char[]			m_current_upload_file;
	
	override TaskId taskId()
	{
		return TaskId.CONTROL;
	}
	
	override void initialize()
	{
	}
	
	override void terminate()
	{
		// nothing
	}
	
	override ReturnCode control()
	{
		ReturnCode 	result;
		uint		rehash_counter = 0;
		
		while (true) {
			result = control_x(rehash_counter);
			if (result != ReturnCode.SUCCESS) {
				break;
			}
			delayTask(10);
		}
		
		return result;
	}
	
	ReturnCode control_x(inout uint rehash_counter)
	{
		Mail mail;
		ReceivedCommand command;
		DownloadItem	item;
		
		// メッセージ待ち
		mail = waitForRequest();
		
		switch (mail.messageId) {
		/* タイマー */
		case MessageId.CLOCK_TICK:
			tick();
			if (++rehash_counter % 30 == 0) {
				environ.downloadList.rehash;
			}
			break;
			
		case MessageId.TERMINATE:
			return ReturnCode.TERMINATE;	/* 終了		*/
			
		/* コマンドルーティング		*/
		case MessageId.ROUTING:
			command = cast(ReceivedCommand)mail.data;
			if (command is null) {
				debug(control_task) mbswritef(derr, "CONTROL: ルーティング失敗\n");
				break;
			}
			try {
				routing(command);
			} catch (Exception e) {
				debug(control_task) mbswritef(derr, "CONTROL: routing exception %s\n", e.toString);
			}
			command = null;
			break;
		/* ダウンロード登録	*/
		case MessageId.ADD_DOWNLOAD:
			item = cast(DownloadItem)mail.data;
			if (item !is null) {
				addDownload(item);
			}
			break;
		/* ダウンロード削除 */
		case MessageId.REMOVE_DOWNLOAD:
			item = cast(DownloadItem)mail.data;
			if (item !is null) {
				removeDownload(item);
			}
			break;
		/* 強制キャッシュ変換 */
		case MessageId.DECODE_CACHE:
			item = cast(DownloadItem)mail.data;
			if (item !is null) {
				decodeCache(item);
			}
			break;
		/* キャッシュ削除 */
		case MessageId.REMOVE_CACHE:
			item = cast(DownloadItem)mail.data;
			if (item !is null) {
				removeCache(item);
			}
			break;
		default:
//				debug(control_task) mbswritef(derr, "Ctrl: 何か受信 %d from %d\n", mail.messageId, mail.from);
			break;
		}
		
		fullCollect();
		
		return ReturnCode.SUCCESS;
	}
	
	void tick()
	{
		workerControl();
	}
	
	void workerControl()
	{
		/* 終了チェック		*/
		
		// エンコーダ
		environ.encoder.signal();
		try {
			foreach (CacheEncoder e; environ.encoder.values) {
				if (e.getState == Thread.TS.TERMINATED) {
					if (e.exitState != 0) {
						debug(control_task) 
							mbswritef(derr, "Encoder: 変換失敗: %s: %s\n",
								m_current_upload_file, e.errorMessage);
					} else {
						debug(control_task) 
							mbswritef(derr,
								"Encoder: 変換成功: ファイル削除 %s\n", m_current_upload_file);
						try {
							forcedRemoveFile(m_current_upload_file);
						} catch (FileException) {
							mbswritef(derr,
								"Encoder: ファイル削除失敗 %s\n", m_current_upload_file);
							
						}
					}
					
					environ.encoder.remove(e.id);
				}
			}
		} finally {
			environ.encoder.release();
		}
		
		// デコーダ
		environ.decoder.signal();
		try {
			foreach (CacheDecoder e; environ.decoder.values) {
				if (e.getState == Thread.TS.TERMINATED) {
					environ.decoder.remove(e.id);
				}
			}
		} finally {
			environ.decoder.release();
		}
		
		// アップローダ
		environ.uploader.signal();
		try {
			foreach (Uploader e; environ.uploader.values) {
				if (e.getState == Thread.TS.TERMINATED) {
					environ.uploader.remove(e.id);
				}
			}
		} finally {
			environ.uploader.release();
		}
		
		// ダウンローダ
		environ.downloader.signal();
		try {
			DownloadItem 	item;
			Distributor		distributor;
			
			foreach (Downloader e; environ.downloader.values) {
				if (e.getState != Thread.TS.TERMINATED) {
					continue;
				}
				distributor	= null;
				item		= null;
				
				environ.downloadList.signal();
				try {
					if (environ.downloadList.exists(e.hash)) {
						item = environ.downloadList.get(e.hash);
					} else {
						errorLog(std.string.format("downloadItem not found: %s", e.hash));
						foreach (_; environ.downloadList.values) {
							errorLog(std.string.format("-- %s", _.hash));
						}
						assert(0); // error
					}
					foreach (Distributor node; item.address) {
						if (compareNode(node.address, e.address)) {
							distributor = node;
						}
					}
					if (distributor is null) {
						assert(0); // error
					}
					distributor.unlock();
					item.endDownload();
					
					// DL数分はプライオリティをあげとく
					distributor.priority = distributor.priority + e.downloadBlocks;
					
					switch (e.exitState) {
					case Downloader.DownloadState.OK:
						distributor.active();
						distributor.priority = distributor.priority + 1;
						item.downloadBlockMaxUp(); // DL可能ブロック数増
						
						// テーブルにある場合、ノードの優先順位を上げる
						environ.nodeTable.downloadSuccess(e.address, config);
						addNode(new NodeDetails(e.address)); // 無いかもしれないので追加する 2度目以降に期待
						
						debug(download) 
							mbswritef(derr, "Downloader: ダウンロード終了 ID:%s 正常\n", e.id);
						break;
						
					case Downloader.DownloadState.BUSY:
						distributor.busy();
						distributor.priority = distributor.priority - 1;
						distributor.waitTime = 10; 
						item.downloadBlockMaxReset(); // DL可能ブロック数リセット
						debug(download) 
							mbswritef(derr, "Downloader: ダウンロード終了 ID:%s ビジー\n", e.id); 
						break;
						
					case Downloader.DownloadState.CONNECT_ERROR:
						distributor.priority = 0;
						distributor.die();
						item.downloadBlockMaxReset(); // DL可能ブロック数リセット
						debug(download) 
							mbswritef(derr, "Downloader: ダウンロード終了 ID:%s 接続エラー\n", e.id); 
						break;
					case Downloader.DownloadState.TIMEOUT:
					default:
						distributor.busy();
						distributor.waitTime = 10;
						distributor.priority = distributor.priority - 0x05;
						if (distributor.priority == 0) {
							debug(download)
								mbswritef(derr, "Downloader: ダウンロード候補削除 ID:%s \n", e.id); 
							
							distributor.die();
						}
						item.downloadBlockMaxReset(); // DL可能ブロック数リセット
						debug(download) {
							if (e.exitState == Downloader.DownloadState.TIMEOUT) {
								mbswritef(derr, "Downloader: ダウンロード終了 ID:%s タイムアウト\n", e.id);
							} else {
								mbswritef(derr, "Downloader: ダウンロード終了 ID:%s 謎のエラー\n", e.id);
							}
						}
						break;
					}
				} finally {
					environ.downloadList.release();
				}
				
				
				environ.downloader.remove(e.id);
			}
		} finally {
			environ.downloader.release();
		}
		
		/* 稼動チェック		*/
		delayTask(5);
		
		// エンコーダ
		environ.encoder.signal();
		try {
			if (environ.encoder.length == 0) { // 1つずつ
				if (! m_upload_files.empty) {
					// アップロードファイルエンコード
					CacheEncoder encoder;
					
					m_current_upload_file 	= m_upload_files.shift;
					encoder					= new CacheEncoder(resource, m_current_upload_file, environ);
					
					encoder.start();
					environ.encoder.set(encoder.id, encoder);
				}
				else {
					scanUploadFolder();
				}
			}
		} finally {
			environ.encoder.release();
		}
		
		// デコーダー
		environ.downloadList.signal();
		environ.decoder.signal();
		try {
			foreach (DownloadItem e; environ.downloadList.values) {
				if (e.blockCount == e.blocks
					&& !e.isDownloading
					&& !e.isComplete)
				{
					downloadComplete(e);
					if (environ.decoder.length == 0) {
						// ダウンロード完了ファイルでコード
						CacheDecoder decoder;
						
						decoder = new CacheDecoder(resource, e.hash, e.dataSize, environ);
						decoder.start();
						environ.decoder.set(decoder.id, decoder);
					} else {
						m_decode_items.push(e);
					}
				}
			}
			if (environ.decoder.length == 0 
				&& !m_decode_items.empty) 
			{
				// ダウンロード完了ファイルデコード
				CacheDecoder 	decoder;
				DownloadItem	item = m_decode_items.shift;
				
				decoder = new CacheDecoder(resource, item.hash, item.dataSize, environ);
				decoder.start();
				environ.decoder.set(decoder.id, decoder);
			}
		} finally {
			environ.downloadList.release();
			environ.decoder.release();
		}
		
		// アップローダー
		environ.uploader.signal();
		try {
//			uint uploaders = environ.uploader.length;
//			if (uploaders) {
//				uint rate = config.speed / uploaders;
//				if (rate == 0) {
//					rate = 1;
//				}
//				foreach (Uploader e; environ.uploader.values) {
//					e.rate = rate; // アップロード速度制御
//				}
//			}
		} finally {
			environ.uploader.release();
		}
		
		// ダウンローダ
		if (environ.downloader.length < config.downloaderMax + environ.uploader.length) {
			environ.downloader.signal();
			environ.downloadList.signal();
			environ.cacheHeader.signal();
			try {
				CacheBitmap 	bitmap;
				DownloadItem 	item;
				Distributor		distributor;
				ulong[][]		fragments;
				
				// DL候補選択
				foreach (DownloadItem e; environ.downloadList.values.sort) {
					if (e.isComplete) {
						continue;
					}
					foreach (Distributor node; e.address.sort) {
						if (node.isActive && !node.isLock) {
							bool found = false;
							foreach (Downloader downloader; environ.downloader.values) {
								if (compareNode(downloader.address, node.address)) {
									found = true;
									break;
								}
							}
							if (! found) {
								item 		= e;
								distributor	= node;
								break;
							}
						}
					}
				}
				if (item !is null) {
					// BITMAPがある場合は取得
					if (environ.cacheHeader.exists(item.hash, item.dataSize)) {
						CacheHeader header = environ.cacheHeader.get(item.hash, item.dataSize);
						bitmap = header.bitmap;
					}
					
					// ダウンロード中のブロックがある場合除く
					foreach (Downloader downloader; environ.downloader.values) {
						DownloadFile dl = downloader.requestFile;
						
						if (dl.hash == item.hash
							&& dl.fileSize == item.dataSize)
						{
							for (uint no = dl.beginBlock; no < dl.beginBlock + dl.count; ++no) {
								bitmap.set(no); // BIT ON
							}
							break;
						}
					}
					fragments = bitmap.fragments(item.dataSize);
					
					if (fragments.length > 0) {
						Downloader downloader;
						
						uint begin, end;
						
						debug(download) {
							mbswritef(derr, "Downloader: データサイズ: %s [%s-%s]\n",
										item.dataSize, fragments[0][0], fragments[0][1]);
							
						}
						begin 	= cast(uint)fragments[0][0];
						end		= cast(uint)(fragments[0][1] - begin);
						
						if (end > item.downloadBlockMax) {
							end = item.downloadBlockMax;
						}
						
						distributor.lock();
						
						downloader = new Downloader(resource, distributor.address,
										new DownloadFile(item.hash, item.dataSize, begin, end), item, environ);
						
						item.tryDownload();
						downloader.start();
						environ.downloader.set(downloader.id, downloader);
						debug(download) 
							mbswritef(derr, "Downloader: ダウンロード開始 ID:%s %s[%s]\n", downloader.id, 
							item.name, distributor.address.toString); 
					}
				}
			} finally {
				environ.downloader.release();
				environ.downloadList.release();
				environ.cacheHeader.release();
			}
		}
	}
	
	// アップロード
	void upload(ReceivedCommand command)
	{
		Uploader uploader;
		
		if (environ.uploader.length >= config.uploaderMax) {
			try {
				command.connection.connection.sendCommand(new NyConnectedLimitation);
			} catch (SocketException) {
				
			} finally {
				command.connection.unlock();
			}
			
			return; // 終了
		}
		
		NyFileRequest 	request 	= new NyFileRequest(command.command);
		UploadFile 		upfile 		= new UploadFile(request.taskId, request.hash, request.fileSize,
															request.beginBlock, request.requestCount);
		uploader	=  new Uploader(resource, command.connection, upfile, environ);
		uploader.start();
		environ.uploader.set(uploader.id, uploader);
		
		command.command.clearBuffer();
		command.command.destroy();
		delete command.command;
		command.command = null;
	}
	
	// アップロードフォルダチェック
	void scanUploadFolder()
	{
		char[][] file_list;
		
		file_list = std.file.listdir(config.uploadFolder);
		foreach (char[] e; file_list) {
			char[] path = config.uploadFolder ~ std.path.sep ~ e;
			if (std.file.isfile(path)) {
				m_upload_files.push(path); // アップロード予定ファイルpush
			}
		}
	}
	
	void routing(ReceivedCommand command)
	{
		switch (command.command.commandCode) {
		case NyCommandCode.QUERY:
			routingQuery(command.node, command.command);
			break;
		case NyCommandCode.DIFFUSION_REQUEST:
			routingDiffusionRequest(command);
			break;
		case NyCommandCode.ANOTHER_NODE:
			routingNodeDetails(command.command);
			break;
		case NyCommandCode.CONDITIONAL_DIFFUSION_REQUEST:
			routingConditionalDiffusionRequest(command);
			break;
		case NyCommandCode.FILE_REQUEST:
			upload(command);
			break;
		default:
//			writef("%s\n", command.command.toString);
			break;
		}
	}
	
	void routingQuery(NodeDetails node, NyRawCommand command)
	{
		NyQuery				query 		= new NyQuery(command);
		NyKeyInformation[]	keys		= new NyKeyInformation[query.keyInformation.length];
		uint i = 0;
		
		command.clearBuffer();
		command.destroy();
		delete command;
		command = null;
		
		foreach (NyKeyInformation e; query.keyInformation) {
			NyKeyInformation key = new NyKeyInformation;
			key.hash				= e.hash.dup;
			key.fileSize			= e.fileSize;
			key.fileName			= e.fileName.dup;
			key.trip				= e.trip.dup;
			key.timer				= e.timer;
			key.fileNameMBSLength	= e.fileNameMBSLength;
			key.keyAddress 			= e.keyAddress;
			key.keyPort				= e.keyPort;
			key.bbsAddress			= e.bbsAddress;
			key.bbsPort				= e.bbsPort;
			key.bbsTrip				= e.bbsTrip.dup;
			key.referredBlock		= e.referredBlock;
			key.modifiedTime		= e.modifiedTime;
			key.ignore				= e.ignore;
			key.keyVersion			= e.keyVersion;
			
			keys[i++] = key;
		}
		
		// クエリタスクへ送信
		sendRequestTo(MailId.QUERY_REQUEST, Mail.makeRequest(MessageId.RECEIVE_QUERY, query));
		
		if (keys.length != 0) {
			if (node.type == NodeDetails.NodeType.NAT) {
				// NATからのキーはアドレスを書き換える
				foreach (NyKeyInformation e; keys) {
					if (e.keyAddress == node.reportAddress.addr) {
						e.keyAddress = node.inetAddress.addr;
					}
				}
			}
			
			/* 保持ノード数が少ない場合は、キーからノード情報を取得する	*/
			if (environ.nodeTable.length < config.nodeMax * 0.025) {
				NodeDetails node;
				
				foreach (NyKeyInformation e; keys) {
					node = new NodeDetails(e.address);
					node.reportAddress(node.inetAddress);
					
					sendRequestTo(MailId.NODE_REQUEST,
						Mail.makeRequest(MessageId.ADD_NODE, node));
				}
			}
			
			// ダウンロード用アドレスを取得
			foreach (NyKeyInformation key; keys) {
				foreach (DownloadItem item; environ.downloadList.values) {
					if (item.isComplete) {
						continue;
					}
					if (key.hash == item.hash && key.fileSize == item.dataSize) {
						Distributor 	distributor;
						InternetAddress	address = key.address;
						
						foreach (DownloadItem e; environ.downloadList.values) {
							if (e.addressExists(address)) {
								distributor = e.distributor(address);
								break;
							}
						}
						if (distributor is null) {
							item.addAddress(address);
						} else {
							item.addAddress(distributor);
						}
					}
				}
			}
			
			// キータスクへ送信
			sendRequestTo(MailId.KEY_REQUEST, 
				Mail.makeRequest(MessageId.ADD_KEY, new KeyInformationArray(keys)));
		}
		query = null;
		keys = null;
	}
	
	void routingDiffusionRequest(ReceivedCommand command)
	{
		sendRequestTo(MailId.QUERY_REQUEST,
			Mail.makeRequest(MessageId.RECEIVE_DIFFUSION_QUERY, command));
	}
	
	void routingConditionalDiffusionRequest(ReceivedCommand command)
	{
		sendRequestTo(MailId.QUERY_REQUEST,
			Mail.makeRequest(MessageId.RECEIVE_COND_DIFFUSION_QUERY, command));
	}
	
	void routingNodeDetails(NyRawCommand command)
	{
		NyAnotherNode	oth	= new NyAnotherNode(command);
		NodeDetails node	= new NodeDetails(new InternetAddress(oth.address, oth.fileSharingPort));
		
		node.speed			= oth.speed;
		node.clusteringKey	= oth.clusteringKey;
		node.correlation 	= clusterCorrelationPoint(config.clusteringKey, node.clusteringKey);
		
		node.reportAddress(node.inetAddress);
		addNode(node);
		
		command.clearBuffer();
		command.destroy();
		delete command;
		command = null;
	}
	
	void addNode(NodeDetails node)
	{
		sendRequestTo(MailId.NODE_REQUEST,
				Mail.makeRequest(MessageId.ADD_NODE, node));
	}
	
	// ダウンローダ停止
	void stopDownloader(DownloadItem item)
	{
		// ダウンロード中の場合は中断する
		environ.downloader.signal();
		try {
			foreach (Downloader e; environ.downloader.values) {
				if (e.hash == item.hash) {
					// 強制中断 & 削除 (Remark: ここで数秒ブロックする可能性あり)
					environ.downloader.remove(e.id);
				}
			}
		} finally {
			environ.downloader.release();
		}
	}
	
	// アップローダ停止
	void stopUploader(DownloadItem item)
	{
		// アップロード中の場合は中断する
		environ.uploader.signal();
		try {
			foreach (Uploader e; environ.uploader.values) {
				UploadFile ul = e.requestFile;
				
				if (ul.hash == item.hash
					&& ul.dataSize == item.dataSize)
				{
					// 強制中断 & 削除 (Remark: ここで数秒ブロックする可能性あり)
					environ.uploader.remove(e.id);
				}
			}
		} finally {
			environ.uploader.release();
		}
	}
	
	void stopDecoder(ubyte[] hash)
	{
		// アップロード中の場合は中断する
		environ.decoder.signal();
		try {
			foreach (CacheDecoder e; environ.decoder.values) {
				if (e.hash == hash) {
					// 強制中断 & 削除 (Remark: ここで数秒ブロックする可能性あり)
					environ.decoder.remove(e.id);
				}
			}
		} finally {
			environ.decoder.release();
		}
	}
	
	void stopEncoder(ubyte[] hash)
	{
		// アップロード中の場合は中断する
		environ.encoder.signal();
		try {
			foreach (CacheEncoder e; environ.encoder.values) {
				if (e.hash == hash) {
					// 強制中断 & 削除 (Remark: ここで数秒ブロックする可能性あり)
					environ.encoder.remove(e.id);
				}
			}
		} finally {
			environ.encoder.release();
		}
	}
	
	void removeDownload(DownloadItem item)
	{
		// ダウンロード中の場合は中断する
		stopDownloader(item);
		
		// リスト削除
		environ.downloadList.signal();
		try {
			if (environ.downloadList.exists(item.hash)) {
				environ.downloadList.remove(item.hash);
			}
		} finally {
			environ.downloadList.release();
		}
		// クエリ削除
		removeSearchKey(std.string.format("%%%s", binaryToHexString(item.hash)));
	}
	
	// キャッシュ削除
	void removeCache(DownloadItem item)
	{
		char[] cache_path;
		
		// 削除キャッシュを操作しているワーカーを
		// すべて止める
		environ.downloader.signal();
		environ.uploader.signal();
		environ.decoder.signal();
		environ.encoder.signal();
		environ.cacheHeader.signal();
		
		try {
			stopDownloader(item);
			stopUploader(item);
			stopDecoder(item.hash);
			stopEncoder(item.hash);
			
			if (environ.cacheHeader.exists(item.hash, item.dataSize)) {
				environ.cacheHeader.remove(item.hash, item.dataSize);
			}
			// キャッシュヘッダー削除
			cache_path = cacheResolve(item.hash, config);
			if (std.file.exists(cache_path)) {
				try {
					// キャッシュファイル削除
					forcedRemoveFile(cache_path);
				} catch (FileException) {
					// do nothing
				}
			}
			
		} finally {
			environ.downloader.release();
			environ.uploader.release();
			environ.decoder.release();
			environ.encoder.release();
			environ.cacheHeader.release();
		}
	}
	
	void decodeCache(DownloadItem item)
	{
		m_decode_items.push(item);
	}
	
	void addDownload(DownloadItem item)
	{
		NyKeyInformation[] 	hit;
		CacheHeader			cache;
		uint before;
		
		if (environ.downloadList.exists(item.hash)) {
			return; // 登録済み
		}
		
		// リストへ登録
		// キーテーブルからサーチ
		hit = environ.keyTable.searchHash(item.hash, 1);
		if (hit.length != 0) {
			item.addAddress(hit[0].address);
		}
		
		environ.cacheHeader.signal();
		try {
			if (environ.cacheHeader.exists(item.hash, item.dataSize)) {
				// ブロック数取得
				cache = environ.cacheHeader.get(item.hash, item.dataSize);
				item.mergeBitmap(cache.bitmap);
				// ファイル名更新
				if (item.name.length > 0 && cache.fileName != item.name) {
					cache.fileName	= item.name;
					environ.cacheHeader.update(cache);
				}
			}
		} finally {
			environ.cacheHeader.release();
		}
		
		environ.downloadList.set(item.hash, item);
		
		// ハッシュ値をクエリ登録
		addSearchKey(std.string.format("%%%s", binaryToHexString(item.hash)));
	}
	
	void downloadComplete(DownloadItem item)
	{
		CacheHeader header;
		
		removeSearchKey(std.string.format("%%%s", binaryToHexString(item.hash)));
		
		environ.downloadList.signal();
		environ.cacheHeader.signal();
		try {
			if (environ.downloadList.exists(item.hash)) {
				item.complete();
				item.clearAddress();
				
				header = environ.cacheHeader.get(item.hash, item.dataSize);
				header.completed = 1;
				environ.cacheHeader.update(header);
			}
		} finally {
			environ.downloadList.release();
			environ.cacheHeader.release();
		}
	}
	
	void removeSearchKey(char[] key)
	{
		// クエリタスクから削除
		sendRequestTo(MailId.QUERY_REQUEST, 
			Mail.makeRequest(MessageId.REMOVE_SEARCH_WORD, new ArrayBox!(char)(key)));
	}
	
	void addSearchKey(char[] key)
	{
		sendRequestTo(MailId.QUERY_REQUEST, 
				Mail.makeRequest(MessageId.ADD_SEARCH_WORD, new ArrayBox!(char)(key)));
	}
	
	
public:
	this(Environment environ, TaskResource resource)
	{
		super(MailId.CONTROL_REQUEST,
				MailId.CONTROL_RESPONSE, environ, resource);
		
		m_upload_files = new Queue!(char[]);
		m_decode_items = new Queue!(DownloadItem);
		
		// ワーカー監視タイマー 1sec
		setMessageTimer(this, 1, MessageTimer.TTL.ANY);
	}
}

