﻿module common.connectiontable;

// コネクションテーブル

private import std.c.time;
private import std.socket;
private import pes.net.winny.all;
private import pes.net.util;
private import pes.system.all;
//private import pes.prompt;
private import common.nodetable;
private import common.configuration;

// コネクション要素
class Connection
{
protected:
	NyConnection		m_connection;
	time_t				m_time;
	NodeInformation		m_node;
	bool				m_lock;
	
public:
	// 能動: 上流検索リンク用
	// 新規接続する
	// テーブルに入れるかはその後の話
	NyProtocolHeader connect(Configuration config, NodeInformation node,
						NyConnectionType.LinkType link_type = NyConnectionType.LinkType.SEARCH)
	{
		NyProtocolHeader 	header;
		NySpeed				speed;
		NyConnectionType	type;
		NyNodeInformation	info;
		
		close();	// 一度クローズしとく
		
		m_connection = new NyConnection;
		m_connection.timeout(config.timeout, 0);
		
		// negotiation
		try {
			header = m_connection.connect(node.inetAddress);
			sendNodeInformation(m_connection, config, link_type);
			receiveNodeInformation(m_connection, speed, type, info);
			node.know(config, node.inetAddress, header, speed, type, info);
		} catch (Exception e) {
			close();
			// クローズだけするので例外は上でキャッチする!!
			// つまりコネクトに失敗した場合は、クローズされている!!
			throw e; // 再送
		}
		
		m_node 	= node;
		m_time	= time(null);
		
		return header;
	}
	
	// 受動: 下流検索リンク、転送リンク用
	// connは、コネクトしてきた接続を渡す
	NodeInformation accept(Configuration config, NyConnection conn,
							NyProtocolHeader header, out NyConnectionType type)
	{
		NySpeed				speed;
		NyNodeInformation	info;
		
		close();	// 一度クローズしとく
		
		NodeInformation node = new NodeInformation(conn.address);
		
		m_connection 	= conn;
		
		try {
			receiveNodeInformation(m_connection, speed, type, info);
			sendNodeInformation(m_connection, config, type.linkType);// 同じタイプを返してみる
			node.know(config, node.inetAddress, header, speed, type, info);
		} catch (Exception e) {
			close();
			// クローズだけするので例外は上でキャッチする!!
			// つまりにアクセプト失敗した場合は、クローズされている!!
			throw e;
		}
		
		m_node			= node;
		m_time			= time(null);
		
		return node;
	}
	
	void close(NyCommand close_command = null)
	{
		m_node = null;
		
		if (m_connection !is null) {
			try {
				if (close_command is null) {
					close_command = new NyClose;
				}
				m_connection.sendCommand(close_command);
			} finally {
				m_connection.close();
				m_connection = null;
			}
		}
	}
	
	// ノード情報送信
	static void sendNodeInformation(NyConnection conn, Configuration config,
									NyConnectionType.LinkType link_type)
	{
		NySpeed 			speed	= new NySpeed;
		NyConnectionType	type	= new NyConnectionType;
		NyNodeInformation	info	= new NyNodeInformation;
		InternetAddress		myip	= getSystemIPAddress(config.fileSharingPort);
		
		speed.speed 	= config.speed;
		type.port0	 	= config.isPort0;
		type.linkType	= link_type;
		
		info.address 	= myip.addr;
		info.port 		= myip.port;
		
		info.clusteringKey(config.clusteringKey);
		
		conn.sendCommand(speed);
		conn.sendCommand(type);
		conn.sendCommand(info);
	}
	
	// ノード情報受信
	static void receiveNodeInformation(NyConnection conn, 
							out NySpeed speed, out NyConnectionType type,
							out NyNodeInformation info)
	{
		int processing = 0;
		
	_RECEIVE_LOOP:
		do {
			NyRawCommand command =conn.receiveCommand();
			switch (command.commandCode) {
			case NyCommandCode.SPEED:
				speed	= new NySpeed(command);
				processing += 1;
				break;
				
			case NyCommandCode.CONNECTION_TYPE:
				type = new NyConnectionType(command);
				processing += 1;
				break;
			case NyCommandCode.NODE_INFORMATION:
				info = new NyNodeInformation(command);
				processing += 1;
				break;
			default:
				throw new ReceiveError; // テスト接続などもこれ
			}
			//delete command;
		} while (processing != 3);	// BBSノードの場合は多い
		
		if (speed is null 
			|| type is null
			|| info is null)
		{
			throw new ReceiveError;
		}
	}
	
	void lock()
	{
		m_lock = true;
	}
	
	void unlock()
	{
		m_lock = false;
	}
	
	// getter
	
	bool isLock()
	{
		return m_lock;
	}
	
	NyConnection connection()
	in {
		assert(m_time != 0);
	}
	body {
		return m_connection;
	}
	
	NodeInformation nodeInformation()
	in {
		assert(m_time != 0);
	}
	body {
		return m_node;
	}
	
	time_t connectTime()
	in {
		assert(m_time != 0);
	}
	body {
		return m_time;
	}
	
	// 連想配列用のキー. ノードテーブルと同じキーという前提
	// 同じノードへは同じ種類の接続を作らないように気をつけること
	// -> 同じノードへ２つの転送リンクは張らない
	ulong aaKey()
	in {
		assert(m_node !is null);
	}
	body {
		return m_node.aaKey();
	}
}

// コネクションテーブル
class ConnectionTable: Locker!(ulong, Connection)
{
public:
	this(SemaphoreResource resource, semaphore id)
	{
		super(resource, id);
	}
	
	override void set(ulong key, Connection conn, LockerMode mode = LockerMode.AUTO)
	{
		if (mode == LockerMode.AUTO) {
			lock();
		}
		try {
			if (exist(key, LockerMode.NON)) {
				Connection oldconn = get(key, LockerMode.NON);
				if (oldconn is conn) {
					return;
				}
				// 同じコネクションは古いほうを切断しとく
				remove(key, LockerMode.NON);
			}
			super.set(key, conn, LockerMode.NON);
		} finally {
			if (mode == LockerMode.AUTO) {
				unlock();
			}
		}
	}
	
	override void remove(ulong key, LockerMode mode = LockerMode.AUTO)
	{
		if (mode == LockerMode.AUTO) {
			lock();
		}
		
		try {
			if (exist(key, LockerMode.NON)) {
				try {
					get(key, LockerMode.NON).close();
				} catch (SocketException) {
					// ここは例外にしない
				}
			}
			super.remove(key, LockerMode.NON); // バグ発見のため外でやる
		} finally {
			if (mode == LockerMode.AUTO) {
				unlock();
			}
		}
	}
	
	override bool exist(ulong key, LockerMode mode = LockerMode.AUTO)
	{
		return super.exist(key, mode); // なんかエラー出るので入れた
	}
	
	bool exist(uint address, uint port, LockerMode mode = LockerMode.AUTO)
	{
		if (mode == LockerMode.AUTO) {
			lock();
		}
		try {
			foreach (Connection e; values(LockerMode.NON)) {
				if ((e.nodeInformation.inetAddress.addr == address
					|| e.nodeInformation.reportAddress.addr == address)
					&& e.nodeInformation.reportAddress.port == port)
				{
					return true; // ok
				}
			}
		} finally {
			if (mode == LockerMode.AUTO) {
				unlock();
			}
		}
		
		return false;
	}
}



